﻿/*
 * Created by SharpDevelop.
 * User: Malte Schulze
 * Date: 16.10.2017
 * Time: 02:53
 * 
 * To change this template use Tools | Options | Coding | Edit Standard Headers.
 */
using RIMMSqol.genericSettings;
using RimWorld;
using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using Verse;

namespace RIMMSqol.renderers
{
    /// <summary>
    /// Look And Feel from a select dialog as a renderer.
    /// </summary>
    public class FilteredGridListRenderer<T> : ParentRendererWithIteration
    {
        protected IOrderedEnumerable<T> items;
        protected FilteredGridListFilter<T> filter;
        protected float totalOptionsWidth = -1, height, width;
        protected Vector2 scrollPosition;
        protected float horizontalGap = 17f, verticalGap = 2f, childGap = 2f;
        public bool SkipOffscreenChildren { get; set; } = false;



        public FilteredGridListRenderer(IEnumerable<T> items, float width, float height, FilteredGridListFilter<T> filter = null)
        {
            if (items is IOrderedEnumerable<T>) this.items = (IOrderedEnumerable<T>)items;
            else this.items = items.OrderBy(d => 0);
            this.width = width;
            this.height = height;
            this.filter = filter;
        }

        public override float getHeight()
        {
            return height;
        }
        public override float getWidth()
        {
            return width;
        }

        public override void DoComponentContents(Rect inRect, Flow flow)
        {
            Rect viewRect = new Rect(inRect);

            IEnumerable<T> filteredItems;
            if (filter != null)
            {
                filter.DoComponentContents(inRect, flow);
                viewRect.yMin += filter.getHeight();

                filteredItems = filter.FilterAndSort(flow, items);
            }
            else
            {
                filteredItems = items;
            }

            Rect contentRect = new Rect(0f, 0f, totalOptionsWidth > 0 ? totalOptionsWidth : viewRect.width, viewRect.height - QOLMod.HorizontalScrollbarHeight());
            bool skipOffscreen = SkipOffscreenChildren;
            Rect visibleContentRect = new Rect(contentRect.x + this.scrollPosition.x, contentRect.y + this.scrollPosition.y, viewRect.width, viewRect.height);
            Widgets.BeginScrollView(viewRect, ref this.scrollPosition, contentRect, true);

            beginIteration(flow);

            Rect r = new Rect(contentRect.x, contentRect.y, 0, 0);
            float columnStart = r.x, maxColumnWidth = 0;
            foreach (T d in filteredItems)
            {
                advanceIteration(d);

                r.height = 0;
                foreach (BaseRenderer cr in childs)
                {
                    r.height = Mathf.Max(r.height, cr.getHeight());
                }

                if (r.yMax > contentRect.height)
                {
                    r.y = contentRect.yMin;
                    r.x += maxColumnWidth + horizontalGap;
                    columnStart = r.x;
                    maxColumnWidth = 0;
                }

                foreach (BaseRenderer cr in childs)
                {
                    r.width = cr.getWidth();
                    if (!skipOffscreen || r.Overlaps(visibleContentRect))
                    {
                        cr.DoComponentContents(r, flow);
                    }
                    r.x += r.width + childGap;
                }
                maxColumnWidth = Mathf.Max(maxColumnWidth, r.x - columnStart - childGap);
                r.x = columnStart;
                r.y += r.height + verticalGap;
            }
            totalOptionsWidth = columnStart + maxColumnWidth + horizontalGap - contentRect.x;

            endIteration(flow);

            Widgets.EndScrollView();
        }
    }

    public abstract class FilteredGridListFilter<T> : BaseRenderer
    {
        abstract public IEnumerable<T> FilterAndSort(Flow flow, IEnumerable<T> items);
    }

    public class FilteredGridListFilter_ISettingsInstance : FilteredGridListFilter<ISettingsInstance>
    {
        protected IEnumerable<ISettingsInstance> cachedFilteredAndSorted;
        protected bool requiresFiltering = true;
        protected String filterExpression;
        protected float width;

        public override IEnumerable<ISettingsInstance> FilterAndSort(Flow flow, IEnumerable<ISettingsInstance> items)
        {
            if (requiresFiltering)
            {
                cachedFilteredAndSorted = items.Where(setting =>
                {
                    if (filterExpression.NullOrEmpty()) return true;
                    if (filterExpression.Contains("#"))
                    {
                        Def def = ((ISettingsVisualizationSelection)SettingsStorage.getSettingsProperties(setting.getId()).visualization).getBaseObjectFromKey(setting.getKey()) as Def;
                        if (def != null && def.modContentPack != null && def.modContentPack.Name != null)
                        {
                            string[] parts = filterExpression.Split(new char[] { '#' }, 2);
                            if (parts.Length == 2)
                            {
                                return (parts[0].Length == 0 || setting.getLabel().IndexOf(parts[0], StringComparison.OrdinalIgnoreCase) >= 0) &&
                                    (parts[1].Length == 0 || def.modContentPack.Name.IndexOf(parts[1], StringComparison.OrdinalIgnoreCase) >= 0);
                            }
                        }
                    }
                    return setting.getLabel().IndexOf(filterExpression, StringComparison.OrdinalIgnoreCase) >= 0;
                }).ToList();
                requiresFiltering = false;
            }
            return cachedFilteredAndSorted;
        }

        public override float getHeight()
        {
            return 35f;
        }

        public override float getWidth()
        {
            return 200f;
        }

        public override void DoComponentContents(Rect inRect, Flow flow)
        {
            string oldFilterExpression = filterExpression;
            inRect.width = 200f; inRect.height = 30f;
            this.filterExpression = Widgets.TextField(inRect, this.filterExpression);
            if ( Mouse.IsOver(inRect) ) TooltipHandler.TipRegion(inRect, () => "RIMMSQoL.EditSettings.TextFilter".Translate(), 1);
            if (oldFilterExpression != filterExpression) requiresFiltering = true;
        }
    }

    public class FilteredGridListFilter_Summary : FilteredGridListFilter<Pair<ISettingsInstance, Action<Flow>>>
    {
        protected IEnumerable<Pair<ISettingsInstance, Action<Flow>>> cachedFilteredAndSorted;
        protected bool requiresFiltering = true;
        protected String filterExpression;
        protected float width;

        public override IEnumerable<Pair<ISettingsInstance, Action<Flow>>> FilterAndSort(Flow flow, IEnumerable<Pair<ISettingsInstance, Action<Flow>>> items)
        {
            if (requiresFiltering)
            {
                cachedFilteredAndSorted = items.Where(item =>
                {
                    if (!item.First.getActive()) return false;
                    if (filterExpression.NullOrEmpty()) return true;
                    ISettingsInstance setting = item.First;
                    if (filterExpression.Contains("#"))
                    {
                        Def def = ((ISettingsVisualizationSelection)SettingsStorage.getSettingsProperties(setting.getId()).visualization).getBaseObjectFromKey(setting.getKey()) as Def;
                        if (def != null && def.modContentPack != null && def.modContentPack.Name != null)
                        {
                            string[] parts = filterExpression.Split(new char[] { '#' }, 2);
                            if (parts.Length == 2)
                            {
                                return (parts[0].Length == 0 || setting.getLabel().IndexOf(parts[0], StringComparison.OrdinalIgnoreCase) >= 0) &&
                                    (parts[1].Length == 0 || def.modContentPack.Name.IndexOf(parts[1], StringComparison.OrdinalIgnoreCase) >= 0);
                            }
                        }
                    }
                    return setting.getLabel().IndexOf(filterExpression, StringComparison.OrdinalIgnoreCase) >= 0;
                }).ToList();
                requiresFiltering = false;
            }
            return cachedFilteredAndSorted;
        }

        public override float getHeight()
        {
            return 35f;
        }

        public override float getWidth()
        {
            return 200f;
        }

        public override void DoComponentContents(Rect inRect, Flow flow)
        {
            string oldFilterExpression = filterExpression;
            inRect.width = 200f; inRect.height = 30f;
            this.filterExpression = Widgets.TextField(inRect, this.filterExpression);
            if (Mouse.IsOver(inRect)) TooltipHandler.TipRegion(inRect, () => "RIMMSQoL.EditSettings.TextFilter".Translate(), 1);
            if (oldFilterExpression != filterExpression) requiresFiltering = true;
        }
    }

    public class FilteredGridListFilter_Backstory : FilteredGridListFilter<ISettingsInstance>
    {
        protected IEnumerable<ISettingsInstance> cachedFilteredAndSorted;
        protected List<string> skillGainsToOrderBy = new List<string>();
        protected bool requiresFiltering = true, requiresSorting = true, allowAdult = true, allowChild = true;
        protected String filterExpression;
        protected float width;
        protected Pawn pawn;

        public FilteredGridListFilter_Backstory()
        {
            pawn = QOLMod.getDummyPawn();
        }

        public override IEnumerable<ISettingsInstance> FilterAndSort(Flow flow, IEnumerable<ISettingsInstance> items)
        {
            if ( requiresFiltering )
            {
                cachedFilteredAndSorted = items.Select(setting => new Pair<ISettingsInstance,BackstoryDef>(setting,(BackstoryDef)((ISettingsVisualizationSelection)
                    SettingsStorage.getSettingsProperties(setting.getId()).visualization).getBaseObjectFromKey(setting.getKey())))
                    .Where(p=>
                {
                    BackstoryDef bs = p.Second;

                    return bs != null && ((bs.slot == BackstorySlot.Adulthood && allowAdult) || (bs.slot == BackstorySlot.Childhood && allowChild)) &&
                        (bs.title.IndexOf(this.filterExpression, StringComparison.OrdinalIgnoreCase) >= 0 /*|| 
                        bs.FullDescriptionFor(pawn).ToString().IndexOf(this.filterExpression, StringComparison.OrdinalIgnoreCase) >= 0*/);
                }).OrderByDescending(p => {
                    BackstoryDef bs = p.Second;
                    if (!bs.skillGains.Any()) return 0;
                    int gains = 0;
                    foreach (KeyValuePair<SkillDef, int> kvp in bs.skillGains)
                    {
                        if (skillGainsToOrderBy.Contains(kvp.Key.defName)) gains += kvp.Value;
                    }
                    return gains;
                }).ThenBy(p => p.Second.title).Select(p=>p.First).ToList();
                requiresFiltering = requiresSorting = false;
            }
            else if ( requiresSorting )
            {
                cachedFilteredAndSorted = cachedFilteredAndSorted.Select(setting => new Pair<ISettingsInstance, BackstoryDef>(setting, (BackstoryDef)((ISettingsVisualizationSelection)
                    SettingsStorage.getSettingsProperties(setting.getId()).visualization).getBaseObjectFromKey(setting.getKey()))).Where(p=>p.Second != null)
                    .OrderByDescending(p => {
                        BackstoryDef bs = p.Second;
                        if (!bs.skillGains.Any()) return 0;
                        int gains = 0;
                        foreach (KeyValuePair<SkillDef, int> kvp in bs.skillGains)
                        {
                            if (skillGainsToOrderBy.Contains(kvp.Key.defName)) gains += kvp.Value;
                        }
                        return gains;
                    }).ThenBy(p => p.Second.title).Select(p => p.First).ToList();
                requiresSorting = false;
            }
            return cachedFilteredAndSorted;
        }

        public override float getHeight()
        {
            return 35f;
        }

        public override float getWidth()
        {
            return width;
        }

        public override void DoComponentContents(Rect inRect, Flow flow)
        {
            string oldFilterExpression = filterExpression;
            Rect rTextInput = new Rect(inRect.x, inRect.y, 200f, 30f);
            this.filterExpression = Widgets.TextField(rTextInput, this.filterExpression);
            if (Mouse.IsOver(inRect)) TooltipHandler.TipRegion(rTextInput, () => "RIMMSQoL.EditSettings.TextFilter".Translate(), 1);
            if (oldFilterExpression != filterExpression) 
            {
                requiresFiltering = true;
            }

            Color oldColor = GUI.color;
            Rect symbolRect = new Rect(rTextInput.x + rTextInput.width + 8f, rTextInput.y, 30f, 30f);
            foreach (SkillDef d in DefDatabase<SkillDef>.AllDefsListForReading)
            {
                Texture2D tex;
                if (!Materials.IconForSkill.TryGetValue(d.defName, out tex))
                {
                    tex = Materials.iconEdit;
                }
                if (Widgets.ButtonImage(symbolRect, tex, skillGainsToOrderBy.Contains(d.defName) ? Color.green : Color.white))
                {
                    if (!skillGainsToOrderBy.Remove(d.defName))
                        skillGainsToOrderBy.Add(d.defName);
                    requiresSorting = true;
                }
                symbolRect.x += symbolRect.width + 8f;
            }
            GUI.color = oldColor;

            string label = "C";
            symbolRect.width += Text.CalcSize(label).x;
            Widgets.Label(symbolRect, label);
            if (Mouse.IsOver(symbolRect)) TooltipHandler.TipRegion(inRect, () => "Childhood".Translate(), 1);
            symbolRect.x += symbolRect.width;
            bool oldValue = allowChild;
            Widgets.Checkbox(new Vector2(symbolRect.x, symbolRect.y), ref allowChild, 24, false);
            if (oldValue != allowChild) requiresFiltering = true;
            symbolRect.x += 24 + 8f;

            label = "A";
            symbolRect.width += Text.CalcSize(label).x;
            Widgets.Label(symbolRect, label);
            if (Mouse.IsOver(symbolRect)) TooltipHandler.TipRegion(inRect, () => "Adulthood".Translate(), 1);
            symbolRect.x += symbolRect.width;
            oldValue = allowAdult;
            Widgets.Checkbox(new Vector2(symbolRect.x, symbolRect.y), ref allowAdult, 24, false);
            if (oldValue != allowAdult) requiresFiltering = true;
            symbolRect.x += 24 + 8f;

            width = symbolRect.x;
        }
    }
}
